# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/2/25
import os
import abc
import copy
import inspect
import datetime
from flask import (
    Flask,
    Blueprint,
    request,
    views,
    jsonify,
)
from flask.json import JSONEncoder
from yqn_project_cli.utils import (
    merge_iterable_objs,
    load_module_from_pyfile,
)
from yqn_project_cli.utils.core.exceptions import APIException
from yqn_project_cli.utils.core.parser import DAJsonBase
from yqn_project_cli.utils.core.logger import Logger
from yqn_project_cli.rpc import RPCConfigBase


class UrlBase:
    def __init__(self, rule, view_func, title, endpoint, **options):
        """
        :param rule:
        :param view_func:
        :param title: url desc
        :param options:
        """
        self.rule = rule
        self.title = title
        self.params = dict(
            view_func=view_func,
            endpoint=endpoint,
            **options
        )

    def __str__(self):
        return str(self.title) + ": " + str(self.rule)


class BlueprintBase(Blueprint):
    def __init__(self, name, import_name, urls, title, **options):
        """
        :param name:
        :param import_name:
        :param urls: list-typed
        :param title: blueprint desc
        :param options:
        """
        super().__init__(name, import_name, **options)
        self.params = dict(
            name=name,
            import_name=import_name,
            **options
        )
        self.title = title
        self.urls = urls
        self.add_url_rules(urls)

    def add_url_rules(self, urls):
        for url in urls:
            self.add_url_rule(url.rule, **url.params)

    def __str__(self):
        return str(self.title) + ": " + str(self.urls)


class DAJson(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")

        elif isinstance(obj, datetime.date):
            return obj.strftime('%Y-%m-%d')

        elif isinstance(obj, bytes):
            return obj.decode("utf-8")

        elif isinstance(obj, DAJsonBase):
            return obj.to_json()

        else:
            return str(obj)

        # else:
        #     return super(DAJson, self).default(o)


class DAFlask(Flask):
    json_encoder = DAJson

    def __init__(self, import_name, project_config, config_object,
                 logger_config=None,
                 **options):

        super(DAFlask, self).__init__(import_name, **options)

        self.project_config = project_config

        self.config.from_object(config_object)

        self.loggers = Logger(logger_config)

        self.plugins = {}

        self.services = {}

        self.server_manager = None  # (server_name, server_connector)

    def add_url_rules(self, rules, endpoint=None, view_func=None, provide_automatic_options=None, **options):
        assert isinstance(rules, (str, list, tuple)), '%s rules allow single str rule or rules list-typed' % rules
        if isinstance(rules, str):
            self.add_url_rule(rules, endpoint=endpoint, view_func=view_func,
                              provide_automatic_options=provide_automatic_options, **options)

        elif isinstance(rules, (list, tuple)):
            for rule in rules:
                self.add_url_rule(rule, endpoint=endpoint, view_func=view_func,
                                  provide_automatic_options=provide_automatic_options, **options)

    def register_urls(self, *urls):
        for url in merge_iterable_objs(urls):
            assert isinstance(url, UrlBase), '%s is not subclass of UrlBase' % type(url)
            self.add_url_rule(url.rule, **url.params)

    def register_blueprints(self, *bps):
        for bp in merge_iterable_objs(bps):
            assert isinstance(bp, BlueprintBase), '%s is not subclass of BlueprintBase' % type(bp)
            self.register_blueprint(bp, **bp.params)

    def register_before_first_requests(self, *fs):
        for f in fs:
            self.before_first_request(f)

    def register_before_requests(self, *fs):
        for f in fs:
            self.before_request(f)

    def register_after_requests(self, *fs):
        for f in fs:
            self.after_request(f)

    def register_teardown_requests(self, *fs):
        for f in fs:
            self.teardown_request(f)

    def register_request_hooks(self, *request_hooks):
        for hook, f in request_hooks:
            if hook == 'before_first_request':
                self.before_first_request(f)

            elif hook == 'before_request':
                self.before_request(f)

            elif hook == 'after_request':
                self.after_request(f)

            elif hook == 'teardown_request':
                self.teardown_request(f)

    def register_error_handlers(self, *error_handlers):
        for code_or_exception, f in error_handlers:
            self.register_error_handler(code_or_exception, f)

    def register_plugins(self, *plugins):
        for _ in plugins:
            assert isinstance(_, Plugin), '%s is not subclass of Plugin' % type(_)
            self.plugins[_.name] = _.plugin(self, *_.args, **_.kwargs)

    def register_plugin_tails(self, *plugin_tails):
        for _ in plugin_tails:
            assert isinstance(_, PluginTail), '%s is not subclass of PluginTail' % type(_)
            _.caller(self.plugins[_.plugin_name], *_.args, **_.kwargs)

    def register_services(self, *servers):
        for _ in servers:
            assert isinstance(_, Server), '%s is not subclass of Server' % type(_)

            if self.server_manager is None and _.as_server_manager:
                server = _.server_cls(client_config=_.client_config,
                                      project_config=copy.deepcopy(self.project_config), **_.kwargs)

                print(server.connect())
                self.server_manager = (_.name, server.connector)

            else:
                server = _.server_cls(client_config=_.client_config,
                                      use_conf_management=_.use_conf_management,
                                      server_manager=self.server_manager[1] if self.server_manager else None,
                                      **_.kwargs)
                server.connect()

            self.services[_.name] = server

        else:
            if self.server_manager is not None:
                setattr(self.services[self.server_manager[0]], 'services', self.services)


class View(views.MethodView):
    http_method_names = ['get', 'post', 'put', 'patch', 'delete', 'head', 'options', 'trace']

    def __init__(self, *args, **kwargs):
        self.args = args
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def as_view(cls, *args, **kwargs):

        for key in kwargs:
            if str(key).lower() in cls.http_method_names:
                raise TypeError("Forbidden http verbs keyword %s to method %s()" % (key, cls.__name__))

            if not hasattr(cls, key):
                raise TypeError("%s() refuse keyword %s which not in attrs-classed" % (cls.__name__, key))

        name = str(args and args[0] or cls.__name__)
        args = args and args[1:]

        return super(View, cls).as_view(name, *args, **kwargs)

    def dispatch_request(self, *args, **kwargs):
        method = request.method.lower() if request.method != "HEAD" else 'get'

        handler = getattr(self, method, self._not_allowed_method)

        return handler(*args, **kwargs)

    def _not_allowed_method(self, *args, **kwargs):
        raise APIException(errmsg='%s is not allowed' % request.method, errcode="-1",
                           data={'methods': [m.upper() for m in self.http_method_names if hasattr(self, m)]})


class RoutePluginBase(metaclass=abc.ABCMeta):
    def __init__(self, app, **kwargs):
        self.app = app
        self.path = os.path.dirname(inspect.getabsfile(self.__class__))
        self.routes = []

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.plugin_app()

    @abc.abstractmethod
    def plugin_app(self):
        raise NotImplementedError

    def auto_register_blueprints(self):
        for pkg in os.listdir(self.path):
            if pkg.startswith(('_', '.')) \
                    or not os.path.isdir(os.path.join(self.path, pkg)) \
                    or not os.path.exists(os.path.join(self.path, pkg, 'route.py')):
                continue

            route = load_module_from_pyfile(os.path.join(self.path, pkg, 'route.py'))

            self.app.register_blueprints(getattr(route, 'blueprints', []))

            self.routes.append(route)


class _JSONResponse:
    def __call__(self, data=None, errmsg='success', errcode='0'):
        return jsonify({'errmsg': errmsg,
                        'errcode': errcode,
                        'data': data})


JSONResponse = _JSONResponse()


class Plugin:
    def __init__(self, name, plugin, *args, **kwargs):
        self.name = name
        self.plugin = plugin
        self.args = args
        self.kwargs = kwargs


class PluginTail:
    def __init__(self, plugin_name, caller, *args, **kwargs):
        """
        caller(plugin, *args, **kwargs)
        """
        self.plugin_name = plugin_name
        self.caller = caller
        self.args = args
        self.kwargs = kwargs


class Server:
    def __init__(self, name, server_cls,
                 client_config=None, use_conf_management=False, as_server_manager=False, **kwargs):
        assert issubclass(server_cls, RPCConfigBase), '%s is not subclass of RPCConfigBase' % type(server_cls)

        self.name = name
        self.server_cls = server_cls
        self.client_config = client_config
        self.use_conf_management = use_conf_management
        self.as_server_manager = as_server_manager
        self.kwargs = kwargs


if __name__ == '__main__':
    print('in as_flask')



# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/3/16
import os
import sys
import json
import shutil
import datetime
import argparse
from string import Template
from termcolor import colored
from yqn_project_cli.utils.handler import auto_view

regen_file_dirs = {
    'project': [
        '$gitignore',
        '$dockerignore',
        '__init__$py',
        'Dockerfile',
        'entrypoint$sh',
        'gunicorn_config$py',
        'gunicorn_config_test$py',
        'model_url_config_local$yml',
        'model_url_config_online$yml',
        'project_config$py',
        'project_config_flask$py',
        'requirements$txt',
        'template_flask$py',
        'template_flask_local$py',
    ],
    'base_dir_files': [
        '$template_config$py',
    ],
    'scheme_dir_files': [
        'config/__init__.py',
    ],
}

pure_schema_dirs = {
    'project': [
        'api',
        'config',
        'rpc',
        'scripts',
        'thirds',
        'utils',
        'outputs',
        'resources',
    ],
    'module': '*',
}

template_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "template")
project_dir = os.path.join(template_dir, 'project')


# project
def output_gen_files(config):
    project_name, app_id, output_dir = config['app_name'], config['app_id'], config['app_path']
    today = datetime.datetime.now().strftime('%Y%m%d')
    for sub_dir in regen_file_dirs:
        files = regen_file_dirs[sub_dir]
        if sub_dir == 'scheme_dir_files':
            for file in files:
                lines = []
                file_path = os.path.join(output_dir, project_name, file)
                regen_file_r = open(file_path, "r")
                tmp = Template(regen_file_r.read())
                regen_file_r.close()
                regen_file_w = open(file_path, "w")
                lines.append(tmp.safe_substitute(project_name=project_name, today=today, app_id=app_id))
                regen_file_w.writelines(lines)
                regen_file_w.close()

        else:
            for file in files:
                if sub_dir == 'base_dir_files':
                    template_file_path = os.path.join(template_dir, file)
                    target_dir = output_dir

                else:
                    template_file_path = os.path.join(template_dir, sub_dir, file)
                    target_dir = os.path.join(output_dir, project_name)

                lines = []
                out_name = file.replace("template_", project_name + "_").replace("$", ".")
                gen_file_path = os.path.join(target_dir, out_name)
                template_file = open(template_file_path)
                gen_file = open(gen_file_path, "w")
                tmp = Template(template_file.read())
                lines.append(tmp.safe_substitute(project_name=project_name, today=today, app_id=app_id))
                gen_file.writelines(lines)
                template_file.close()
                gen_file.close()

    # save source json-file config to project config dir for add or update later
    # and then out rid of source json-file config
    shutil.copyfile(config['src_config'], os.path.join(output_dir, project_name, 'config/{}.json'.format(project_name)))


def create_project(config):
    for _dir in pure_schema_dirs['project']:
        if _dir not in ['outputs', 'resources']:
            source_path = os.path.join(template_dir, 'project', _dir)
            target_path = os.path.join(config['app_path'], config['app_name'], _dir)
            shutil.copytree(source_path, target_path)
        else:
            os.mkdir(os.path.join(config['app_path'], config['app_name'], _dir))

    output_gen_files(config)

    print(colored('%s created successfully at %s' % (config['app_name'], config['app_path']), 'green'))


def init_project(config):
    auto_view(config, init=True)
    print(colored(
        '%s has inited, you can enter %s, and run: python %s_flask_local.py' %
        (config['app_name'], os.path.join(config['app_path'], config['app_name']), config['app_name']), 'green'))


def parse_json(args):
    conf = json.load(open(args.config, 'rb'))

    if not (conf.get('app_name', False)
            and isinstance(conf['app_name'], str)):
        raise ValueError('%s 中参数 app_name 需为有效非空字符串' % args.config)

    if not (conf.get('app_id', False)
            and str(conf['app_id']).isdigit()):
        raise ValueError('%s 中参数 app_id 需为有效数值字符串' % args.config)

    if not (conf.get('app_path', False)
            and isinstance(conf['app_path'], str)
            and os.path.exists(conf['app_path'])
            and conf['app_path'] == os.path.abspath(conf['app_path'])):
        raise ValueError('%s 中参数 app_path 需为有效完整路径字符串' % args.config)

    # save source json-file config to project config dir for add or update later
    # and then out rid of source json-file config
    conf['src_config'] = args.config

    return conf


def parse_command():
    parser = argparse.ArgumentParser(description='more faster to create project')
    parser.add_argument('-c', '--config', type=str, required=True,
                        help='json-file config absolute path, '
                             'you can reference sample from README.md in installed yqn_project_cli')

    args = parser.parse_args()

    param_str = '\n'.join(['%20s = %s' % (k, v) for k, v in sorted(vars(args).items())])

    print('usage: %s\n%20s   %s\n%s\n%s\n' % (' '.join(sys.argv), 'ARG', 'VALUE', '_' * 50, param_str))

    print(colored('refer to', 'green'),
          colored('README.md', 'red'),
          colored('from yqn_project_cli in installed packages for more usage info', 'green'))

    return args


def main():
    conf = parse_json(parse_command())

    create_project(conf)

    init_project(conf)


def auto_handler():
    conf = parse_json(parse_command())

    auto_view(conf)

    print(colored('%s has updated ...' % conf['app_name'], 'green'))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/2/25
from flask import (
    request,
    url_for,
    redirect,
    current_app
)
from yqn_project_cli.utils.core.as_flask import (
    UrlBase,
    BlueprintBase
)
from yqn_project_cli.utils.core.exceptions import PortalException


class PortalUrl(UrlBase):
    def __init__(self, rules, view_func=None, in_menu=True, order=0, title='', icon='', context=None, **options):
        super(PortalUrl, self).__init__(rules, view_func, title, **options)

        self.in_menu = in_menu
        self.order = order
        self.title = title if title else view_func.__name__
        self.icon = icon
        self.context = context
        self.rules = rules,
        self.view_func = view_func,
        self.options = options


class PortalNotMenuUrl(PortalUrl):
    def __init__(self, rules, view_func=None, in_menu=False, order=0, title='', icon='', context=None, **options):
        super(PortalNotMenuUrl, self).__init__(rules, view_func, in_menu, order, title, icon, context, **options)


class PortalBlueprint(BlueprintBase):
    pass


class PortalMenu:
    def __init__(self, portal_blueprints=None, title='', in_menu=True, order=0, icon='', context=None):
        self.portal_blueprints = portal_blueprints
        self.title = title
        self.in_menu = in_menu
        self.order = order
        self.icon = icon
        self.context = context


def portal_child_menus(*portal_blueprints):
    menus = []

    for blueprint in portal_blueprints:

        for url in blueprint.urls:

            if url.in_menu:
                if blueprint.url_prefix is not None:
                    _url = "/".join((blueprint.url_prefix.rstrip("/"), url.rule.lstrip("/")))

                else:
                    _url = url.rule

                menus.append({
                    'order': url.order,
                    'title': url.title,
                    'icon': url.icon,
                    'context': url.context,
                    'url': _url
                })

    return sorted(menus, key=lambda x: x['order'], reverse=True)


def portal_menu_register(*portal_menus):
    show_menus_detail = []

    for menu in [_ for _ in portal_menus if _.in_menu]:
        show_menus_detail.append(
            {
                'title': menu.title,
                'order': menu.order,
                'icon': menu.icon,
                'context': menu.context,
                'menus': portal_child_menus(*menu.portal_blueprints)
            }
        )

    return sorted(show_menus_detail, key=lambda x: x['order'], reverse=True)



# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/2/24
import json
from werkzeug.exceptions import HTTPException


class BasicException(BaseException):
    """for apollo exceptions"""

    def __init__(self, msg: str):
        self._msg = msg
        print(msg)

    def __str__(self):
        return "%s: %s" % (self.__name__, self._msg)


class NameSpaceNotFoundException(BasicException):
    """for apollo exceptions"""


class ServerNotResponseException(BasicException):
    """for apollo exceptions"""


class _Exception(HTTPException):
    errmsg = None
    errcode = None
    data = None
    code = None

    def __init__(self, errmsg=None, errcode=None, data=None, code=None):
        super(_Exception, self).__init__(errmsg, None)

        if errmsg is not None:
            self.errmsg = errmsg

        if errcode is not None:
            self.errcode = errcode

        if data is not None:
            self.data = data

        if code is not None:
            self.code = code

    def get_body(self, environ=None):
        body = dict(
            errcode=self.errcode,
            errmsg=self.errmsg,
            data=self.data
        )
        text = json.dumps(body, sort_keys=False, ensure_ascii=False)

        return text

    def get_headers(self, environ=None):
        return [('Content-Type', 'application/json')]


class APIException(_Exception):
    code = 400
    errmsg = 'sorry, request error'
    errcode = "-1"


class PortalException(_Exception):
    code = 500
    errmsg = 'sorry, internal error'
    errcode = "-1"



# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/3/13
from flask_restx import SchemaModel


class ResponseSchemaBase(SchemaModel):
    def __init__(self, name, schema=None):
        self.name = name
        response_params = self.get_response_params_schema() or schema or {}
        super(ResponseSchemaBase, self).__init__(self.name, response_params)

    def get_response_params_schema(self):
        header = {'type': 'object', 'properties': {}}
        header['properties']['xTraceId'] = {
            'type': 'string',
            'description': "调用链"
        }
        header['properties']['xAppId'] = {
            'type': 'string',
            'description': "调用方app id"
        }
        response_params = {
            "type": "object",
            "properties": {
                'success': {
                    'type': 'boolean',
                    'default': True
                },
                'header': header,
                "data": self.get_model_schema(),
                'msgCode': {
                    'type': 'string',
                },
                'code': {
                    'type': 'integer',
                },
                'msg': {
                    'type': 'string',
                },
                'msgDetail': {
                    'type': 'string',
                },
                'headers': {
                    'type': 'object',
                },
                'bizExtMap': {
                    'type': 'string',
                },
            }
        }
        return response_params

    def get_model_schema(self: dict):
        pass




# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/3/13
from flask_restx import SchemaModel


class RequestSchemaBase(SchemaModel):
    def __init__(self, name, schema=None):
        self.name = name
        request_params = self.get_request_params_schema() or schema or {}
        super(RequestSchemaBase, self).__init__(self.name, request_params)

    def get_request_params_schema(self):
        header = {'type': 'object', 'properties': {}}
        header['properties']['accessToken'] = {
            'type': 'string',
            'description': "accessToken"
        }
        header['properties']['viewAll'] = {
            'type': 'boolean',
            'default': True,
            'description': "是否展示所有"
        }
        header['properties']['xAppId'] = {
            'type': 'string',
            'description': "调用方app id"
        }
        header['properties']['xBPId'] = {
            'type': 'string',
            'description': "xBPId"
        }
        header['properties']['xCallerId'] = {
            'type': 'string',
            'description': "调用方请求ID"
        }
        header['properties']['xClientIp'] = {
            'type': 'string',
            'description': "客户端IP"
        }
        header['properties']['xDeviceId'] = {
            'type': 'string',
            'description': "设备ID"
        }
        header['properties']['xIsTest'] = {
            'type': 'boolean',
            'description': "xIsTest",
            'default': False,
        }
        header['properties']['xJsFinger'] = {
            'type': 'string',
            'description': "Js指纹"
        }
        header['properties']['xLangCode'] = {
            'type': 'string',
            'description': "语言"
        }
        header['properties']['xOpenId'] = {
            'type': 'string',
            'description': "开放ID"
        }
        header['properties']['xOpenPlatform'] = {
            'type': 'integer',
            'default': 0,
            'description': "开放平台"
        }
        header['properties']['xPushToken'] = {
            'type': 'string',
            'description': "app推送token"
        }
        header['properties']['xSession'] = {
            'type': 'string',
            'description': "登录的Session"
        }
        header['properties']['xSourceAppId'] = {
            'type': 'string',
            'description': "初始来源调用方app id"
        }
        header['properties']['xSystemLangCode'] = {
            'type': 'string',
            'description': "系统环境语言"
        }
        header['properties']['xTestFlag'] = {
            'type': 'string',
            'description': "xTestFlag"
        }
        header['properties']['xToken'] = {
            'type': 'string',
            'description': "xToken"
        }
        header['properties']['xTraceId'] = {
            'type': 'string',
            'description': "调用链"
        }
        header['properties']['xUserId'] = {
            'type': 'integer',
            'default': 0,
            'description': "用户ID"
        }
        header['properties']['xUserName'] = {
            'type': 'string',
            'description': "用户名"
        }
        request_params = {
            "type": "object",
            "properties": {
                'header': header,
                "model": self.get_model_schema()
            }
        }
        return request_params

    def get_model_schema(self):
        pass




# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/2/24
import abc


class RPCConfigBase(metaclass=abc.ABCMeta):
    def __init__(self, client_config=None, use_conf_management=False, project_config=None,
                 server_manager=None, **kwargs):
        assert client_config is not None or use_conf_management is True, 'client_config or use_conf_management ?!'

        self.client_config = client_config  # local client config from environ or file
        self.use_conf_management = use_conf_management  # use server config management or not
        self.project_config = project_config  # know project config info from from master or not
        self.server_manager = server_manager  # server conf manager like apollo

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.connector = None

    @abc.abstractmethod
    def connect(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def reconnect(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def close(self, **kwargs):
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/3/11
from yqn_project_cli.utils.core.parser import DAArgument, DARequestParser


class MainIndexParser(DARequestParser):
    choices1 = DAArgument('choices1', required=True, type=int, location='args')
    choices2 = DAArgument('choices2', required=True, type=int, location='args')



# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/3/13
import abc
import inspect
from flask_restx.reqparse import (
    Argument,
    RequestParser
)
from yqn_project_cli.utils.core.exceptions import APIException


class DAArgument(Argument):
    """ http request 参数对象化及校验提示转换 """
    def handle_validation_error(self, error, bundle_errors):

        error_str = str(error)

        if "Missing required parameter" in error_str:
            error_msg = '必传参数'

        elif 'not a valid choice' in error_str:
            error_msg = '不是有效选项'

        elif 'invalid literal for int' in error_str or 'not a valid integer' in error_str:
            error_msg = '不是有效整数'

        elif 'must be a non-negative integer' in error_str:
            error_msg = '不是非负整数'

        elif 'must be a positive integer' in error_str:
            error_msg = '不是正整数'

        elif 'must be within the range' in error_str:
            error_msg = '不在有效范围'

        elif 'boolean type must be non-null' in error_str:
            error_msg = '布尔类型禁止为空'

        elif 'Invalid literal for boolean()' in error_str:
            error_msg = '不支持转为布尔类型'

        elif 'Invalid date literal' in error_str:
            error_msg = '不是有效日期'

        elif 'not a valid ipv4 address' in error_str:
            error_msg = '不是有效IPv4'

        elif 'not a valid ip' in error_str:
            error_msg = '不是有效IP'

        elif 'not a valid email' in error_str:
            error_msg = '不是有效邮箱'

        elif 'does not match pattern' in error_str:
            error_msg = '不符匹配规则'

        elif 'Expected a valid ISO8601 date/time interval' in error_str or 'must be a valid ISO8601 date/time interval' in error_str:
            error_msg = '不是ISO8601格式日期时间'

        elif 'not a valid URL' in error_str:
            error_msg = '不是有效链接'

        else:
            error_msg = error_str

        raise APIException('请求参数有误', data={self.name: error_msg})


class DARequestParser(RequestParser):
    """ http request 参数解析 """
    def __init__(self, *args, **kwargs):
        super(DARequestParser, self).__init__(*args, **kwargs)
        arguments = inspect.getmembers(self.__class__, lambda arg: isinstance(arg, DAArgument))
        for name, argument in arguments:
            self.add_argument(argument)


class DAJsonBase(metaclass=abc.ABCMeta):
    """DAJson解析调用"""
    @abc.abstractmethod
    def to_json(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Author: ZKH
# Date：2021/3/23
from distutils.core import setup

from setuptools import find_packages

with open('./yqn_project_cli/README.md') as reader:
    long_description = reader.read()

if __name__ == '__main__':
    setup(
        name='yqn_project_cli',
        version='0.0.0rc54',
        author='ZouKaihua',
        author_email="zoukaihua@aliyun.com",
        description='For more faster to create semi-project based on flask',
        url='https://blog.zoukaihua.com',
        long_description=long_description,
        long_description_content_type="text/markdown",
        include_package_data=True,
        package_data={'yqn_project_cli': ['template/*', 'template/*/*', 'README.md']},
        packages=find_packages(),
        install_requires=[
            'DBUtils>=2.0',
            'Flask>=1.1.2',
            'Flask-Compress>=1.8.0',
            'Flask-Cors>=3.0.10',
            'flask-restx>=0.2.0',
            'PyMySQL>=1.0.2',
            'pyodps>=0.10.6',
            'requests>=2.25.1',
            'requests-toolbelt>=0.9.1',
            'termcolor>=1.1.0',
        ],
        entry_points={
            'console_scripts': [
                'yqn-project = yqn_project_cli.init_project:main',
                'yqn-autoview = yqn_project_cli.init_project:auto_handler',
            ],
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
    )



import json
import logging
import os
import threading
import time
from telnetlib import Telnet
from typing import Any, Dict, Optional

import requests

from yqn_project_cli.utils.core.exceptions import BasicException, NameSpaceNotFoundException, ServerNotResponseException


class ApolloClient:

    def __new__(cls, *args, **kwargs):
        """
        singleton model
        """
        tmp = {_: kwargs[_] for _ in sorted(kwargs)}
        key = f"{args},{tmp}"
        if hasattr(cls, "_instance"):
            if key not in cls._instance:
                cls._instance[key] = super().__new__(cls)
        else:
            cls._instance = {key: super().__new__(cls)}
        return cls._instance[key]

    def __init__(
            self,
            app_id: str,
            cluster: str = "default",
            config_server_url: str = "http://192.168.10.223:8080",
            env: str = "FAT",
            ip: str = None,
            timeout: int = 10,
            cycle_time: int = 30,
            cache_file_path: str = None,
            authorization: str = None,
    ):
        """
        init method
        :param app_id: application id
        :param cluster: cluster name, default value is 'default'
        :param config_server_url: with the format 'http://localhost:80080'
        :param env: environment, default value is 'DEV'
        :param timeout: http request timeout seconds, default value is 60 seconds
        :param ip: the deploy ip for grey release, default value is the local ip
        :param cycle_time: the cycle time to update configuration content from server
        :param cache_file_path: local cache file store path
        """
        self.config_server_url = config_server_url
        self.app_id = app_id
        self.cluster = cluster
        self.timeout = timeout
        self.stopped = False
        self._env = env
        self.ip = self.init_ip(ip)
        remote = self.config_server_url.split(":")
        self.host = f"{remote[0]}:{remote[1]}"
        if len(remote) == 1:
            self.port = 8090
        else:
            self.port = int(remote[2])
        self._authorization = authorization

        self._request_model = None
        self._cache: Dict = {}
        self._notification_map = {'application': -1}
        # if namespaces is None:
        #     namespaces = ["application"]
        # self._notification_map = {namespace: -1 for namespace in namespaces}
        self._cycle_time = cycle_time
        self._hash: Dict = {}
        if cache_file_path is None:
            self._cache_file_path = os.path.join(
                os.path.abspath(os.path.dirname(__file__)), "config"
            )
        else:
            self._cache_file_path = cache_file_path
        self._path_checker()
        self._stopping = False
        # for http request extension
        # self.start()

    def _get_clusters(self) -> dict:
        """
        get clusters by app id
        :return :
        """
        url = f"{self.host}:{self.port}/apps/{self.app_id}/clusters"
        r = self._http_get(url)
        if r.status_code == 200:
            return r.json()
        else:
            return {}

    def _get_namespaces(self) -> dict:
        """
        get namespaces by app id and cluster
        :return :
        """
        # url = f"{self.host}:{self.port}/apps/{self.app_id}/clusters/{self.cluster}/namespaces"
        # r = self._http_get(url)
        url = '{}/notifications/v2'.format(self.config_server_url)
        notifications = []
        temp = {'application': -1}
        for key in temp:
            notification_id = temp[key]
            notifications.append({
                'namespaceName': key,
                'notificationId': notification_id
            })

        r = requests.get(url=url, params={
            'appId': self.app_id,
            'cluster': self.cluster,
            'notifications': json.dumps(notifications, ensure_ascii=False)
        }, timeout=self.timeout)
        if r.status_code == 200:
            namespaces = r.json()
            return {_.get("namespaceName"): _.get("notificationId") for _ in namespaces}
        else:
            return {"application": -2}

    @staticmethod
    def init_ip(ip: Optional[str]) -> str:
        """
        for grey release
        :param ip:
        :return:
        """
        if ip is None:
            try:
                import socket

                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                s.connect(("114.114.114.114", 53))
                ip = s.getsockname()[0]
                s.close()
            except BaseException:
                return "127.0.0.1"
        return ip

    def get_value(
            self, key: str, default_val: str = None, namespace: str = "application"
    ) -> Any:
        """
        get the configuration value
        :param key:
        :param default_val:
        :param namespace:
        :return:
        """
        try:
            # check the namespace is existed or not
            if namespace in self._cache:
                return self._cache[namespace].get(key, default_val)
            return default_val
        except BasicException:
            return default_val

    def stop(self, ) -> None:
        self._stopping = True

    def start(self, call_back_func) -> None:
        """
        Start the long polling loop.
        :return:
        """
        # check the cache is empty or not
        # if len(self._cache) == 0:
        #     self._long_poll()
        #     call_back_func()
        # start the thread to get config server with schedule
        self._first_request(call_back_func)
        t = threading.Thread(target=self._listener, args=(call_back_func,))
        t.setDaemon(True)
        t.start()

    def _http_get(self, url: str, params: Dict = None) -> requests.Response:
        """
        handle http request with get method
        :param url:
        :return:
        """
        if self._request_model is None:
            return self._request_get(url, params=params)
        else:
            return self._request_model(url)

    def _request_get(self, url: str, params: Dict = None) -> requests.Response:
        """
        :param url:
        :param params:
        :return:
        """
        try:
            if self._authorization:
                return requests.get(
                    url=url,
                    params=params,
                    timeout=self.timeout // 2,
                    headers={"Authorization": self._authorization},
                )
            else:
                return requests.get(url=url, params=params, timeout=self.timeout // 2)

        except requests.exceptions.ReadTimeout:
            # if read timeout, check the server is alive or not
            try:
                tn = Telnet(host=self.host, port=self.port, timeout=self.timeout // 2)
                tn.close()
                # if connect server succeed, raise the exception that namespace not found
                raise NameSpaceNotFoundException("namespace not found")
            except ConnectionRefusedError:
                # if connection refused, raise server not response error
                raise ServerNotResponseException(
                    "server: %s not response" % self.config_server_url
                )

    def _path_checker(self) -> None:
        """
        create configuration cache file directory if not exits
        :return:
        """
        if not os.path.isdir(self._cache_file_path):
            os.mkdir(self._cache_file_path)

    def _update_local_cache(
            self, release_key: str, data: str, namespace: str = "application"
    ) -> None:
        """
        if local cache file exits, update the content
        if local cache file not exits, create a version
        :param release_key:
        :param data: new configuration content
        :param namespace::s
        :return:
        """
        # trans the config map to md5 string, and check it's been updated or not
        if self._hash.get(namespace) != release_key:
            # if it's updated, update the local cache file
            with open(
                    os.path.join(
                        self._cache_file_path,
                        "%s_configuration_%s.txt" % (self.app_id, namespace),
                    ),
                    "w",
            ) as f:
                new_string = json.dumps(data)
                f.write(new_string)
            self._hash[namespace] = release_key

    def _get_local_cache(self, namespace: str = "application") -> Dict:
        """
        get configuration from local cache file
        if local cache file not exits than return empty dict
        :param namespace:
        :return:
        """
        cache_file_path = os.path.join(
            self._cache_file_path, "%s_configuration_%s.txt" % (self.app_id, namespace)
        )
        if os.path.isfile(cache_file_path):
            with open(cache_file_path, "r") as f:
                result = json.loads(f.readline())
            return result
        return {}

    def _get_config_by_namespace(self, namespace: str = "application") -> None:
        """
        :param namespace:
        :return:
        """
        url = '{}/configs/{}/{}/{}?ip={}'.format(self.config_server_url, self.app_id, self.cluster, namespace, self.ip)
        # url = f"{self.host}:{self.port}/apps/{self.app_id}/clusters/{self.cluster}/namespaces/{namespace}/releases/latest"
        try:
            r = self._http_get(url)
            if r.status_code == 200:
                data = r.json()
                self._cache[namespace] = data.get("configurations", "{}")
                self._update_local_cache(
                    data.get("releaseKey", str(time.time())),
                    data.get("configurations", {}),
                    namespace,
                )
            else:
                data = self._get_local_cache(namespace)
                self._cache[namespace] = data
        except BaseException as e:
            logging.getLogger(__name__).warning(str(e))
            data = self._get_local_cache(namespace)
            self._cache[namespace] = data

    def _long_poll(self) -> None:
        try:
            # url = '{}/notifications/v2'.format(self.config_server_url)
            # notifications = []
            # for key in self._notification_map:
            #     notification_id = self._notification_map[key]
            #     notifications.append({
            #         'namespaceName': key,
            #         'notificationId': notification_id
            #     })
            #
            # r = requests.get(url=url, params={
            #     'appId': self.app_id,
            #     'cluster': self.cluster,
            #     'notifications': json.dumps(notifications, ensure_ascii=False)
            # }, timeout=self.timeout)
            new_notification_map = self._get_namespaces()
            if new_notification_map == self._notification_map:
                pass
            else:
                self._notification_map = new_notification_map
                for namespace in self._notification_map.keys():
                    self._get_config_by_namespace(namespace)
        except requests.exceptions.ReadTimeout as e:
            logging.getLogger(__name__).warning(str(e))
        except requests.exceptions.ConnectionError as e:
            logging.getLogger(__name__).warning(str(e))
            self._load_local_cache_file()

    def _load_local_cache_file(self) -> bool:
        """
        load all cached files from local path
        is only used while apollo server is unreachable
        :return:
        """
        for file_name in os.listdir(self._cache_file_path):
            file_path = os.path.join(self._cache_file_path, file_name)
            if os.path.isfile(file_path):
                file_simple_name, file_ext_name = os.path.splitext(file_name)
                if file_ext_name == ".swp":
                    continue
                namespace = file_simple_name.split("_")[-1]
                with open(file_path) as f:
                    self._cache[namespace] = json.loads(f.read())["configurations"]
        return True

    def _first_request(self, call_back_func) -> None:
        self._long_poll()
        call_back_func()

    def _listener(self, call_back_func) -> None:
        """
        :return:
        """
        logging.getLogger(__name__).info("Entering listener loop...")
        while not self._stopping:
            self._long_poll()
            call_back_func()
            time.sleep(self._cycle_time)
